
struct AutoFwdDerivs{R,C}
  hess_res::R
  hess_cfg::C
  hess_buf::Matrix{Float64}
end

function AutoFwdDerivs(fn, n; cs=min(16, n))
  r = DiffResults.HessianResult(zeros(n))
  cfg = ForwardDiff.HessianConfig(fn, r, zeros(n), ForwardDiff.Chunk{cs}())
  AutoFwdDerivs(r, cfg, zeros(n, n))
end

function _gradient(store, fn::F, derivs::AutoFwdDerivs{R,C}, p) where{F,R,C}
  ForwardDiff.gradient!(store, fn, p)
end

function _hessian(store, fn::F, derivs::AutoFwdDerivs{R,C}, p) where{F,R,C}
  res = derivs.hess_res
  cfg = derivs.hess_cfg
  ForwardDiff.hessian!(res, fn, p, cfg)
  store .= DiffResults.hessian(res)
end


struct AutoMixedDerivs{W<:FoRWrapper} 
  wrapper::W
end

function AutoMixedDerivs(fn, n; cs=min(16, n))
  wrapper = FoRWrapper(fn, zeros(n), cs)
  AutoMixedDerivs(wrapper)
end

function _gradient(buf, fn::F, D::AutoMixedDerivs{W}, x) where{F,W} 
  buf .= D.wrapper(:gradient, x)
end

function _hessian(buf, fn::F, D::AutoMixedDerivs{W}, x) where{F,W}
  buf .= D.wrapper(:hessian, x)
end

function _lag_hessian(store, 
                      fn::F, 
                      cons::Constraints{G},
                      derivs::D, 
                      p, o, l) where{F,G,D}
  n    = length(p)
  hbuf = zeros(n, n)
  iszero(o) || _hessian(hbuf, fn, derivs, p)
  hbuf .*= o
  if !all(iszero, l)
    lc = LagConstraintTerm(cons, l)
    hbuf .+= ForwardDiff.hessian(lc, p)
  end
  linear_buffer_ltri_colmajor!(store, hbuf)
  nothing
end

