
function ipopt_optimize(obj, ini, constraints=noconstraints(); 
                        derivs=AutoFwdDerivs(obj, length(ini)),
                        conderivs=AutoFwdConDerivs(length(ini), nconstr(constraints)),
                        jac_nz=length(ini)*nconstr(constraints),
                        box_lower=-1e22, box_upper=1e22, kwargs...)
  # get the dimension of the problem and prepare box constraints:
  arglen = length(ini)
  box_l  = prepare_bounds(arglen, box_lower)
  box_u  = prepare_bounds(arglen, box_upper)
  # create the model object and set up the problem:
  model = NLP(obj, constraints, derivs, conderivs)
  prob  = CreateIpoptProblem(arglen,  box_l, box_u,     # arg count and box
                             constraints.ncon,          # number of constraints
                             constraints.box_lower,     # constraints lower box
                             constraints.box_upper,     # constraints upper box
                             jac_nz,                    # size of constr. jacobian
                             div(arglen*(arglen+1), 2), # size of hessian ltri
                             model,                     # objective
                             ConstraintEval(model),     # constraint function
                             GradientEval(model),       # objective gradient
                             ConstraintJac(model),      # constraint jacobian
                             LagrangianHess(model))     # lagrangian hessian
  # Add all the options:
  for (opt, val) in merge(DEF_IPOPT_KWARGS, Dict(kwargs))
    ipopt_addoption!(prob, opt, val)
  end
  # Solve and return:
  prob.x = copy(ini)
  status = IpoptSolve(prob)
  (status=Ipopt._STATUS_CODES[status], minimizer=copy(prob.x), minval=obj(prob.x))
end

struct DummyZero <: Function end
(dz::DummyZero)(x) = zero(eltype(x))

function ipopt_nlsolve(fn, ini; kwargs...)
  n    = length(ini)
  cons = Constraints(length(ini), fn, zeros(n), zeros(n))
  ipopt_optimize(DummyZero(), ini, cons; kwargs...)
end

