
struct NLP{O,C,D,J} <: Function
  obj::O
  constraints::Constraints{C}
  derivs::D
  conderivs::J
end

# A struct wrapper disambiguates the method call from gradient evaluation.
struct ConstraintEval{P} <: Function
  nlp::P
end

# A struct wrapper disambiguates the method call from constraint evaluation.
struct GradientEval{P} <: Function
  nlp::P
end

struct ConstraintJac{P} <: Function
  nlp::P
end

struct LagrangianHess{P} <: Function
  nlp::P
end

# Lagrangian constraint term (sum of lambda_j*con_j(x)).
struct LagConstraintTerm{J}
  cons::Constraints{J} 
  lam::Vector{Float64}
end

function (lc::LagConstraintTerm{J})(p::AbstractVector{P}) where{J,P}
  buf = zeros(P, lc.cons.ncon)
  lc.cons.fun(buf, p)
  sum(prod, zip(buf, lc.lam))
end

# first method: objective evaluation.
function (nlp::NLP{O,C,D,J})(p) where{O,C,D,J} 
  try
    return nlp.obj(p)
  catch er
    er isa InterruptException && rethrow(er)
    return NaN
  end
end

# second method: constraint evaluation.
function (wrapper::ConstraintEval{NLP{O,C,D,J}})(p, store) where{O,C,D,J} 
  wrapper.nlp.constraints(p, store)
end

# third method: objective gradient.
function (wrapper::GradientEval{NLP{O,C,D,J}})(p, store) where{O,C,D,J} 
  _gradient(store, wrapper.nlp.obj, wrapper.nlp.derivs, p)
end

# fourth method: constraint jacobian.
function (wrapper::ConstraintJac{NLP{O,C,D,J}})(p, r, c, store) where{O,C,D,J} 
  nlp  = wrapper.nlp
  cons = nlp.constraints
  isnothing(store) && return jac_structure!(r, c, cons.ncon, length(p))
  conderivs = nlp.conderivs
  _con_jacobian(store, cons, conderivs, p)
end

# fifth method: lagrangian Hessian.
function (wrapper::LagrangianHess{NLP{O,C,D,J}})(p, r, c, o, l, store) where{O,C,D,J}
  nlp = wrapper.nlp
  isnothing(store) && return hes_structure!(r, c, length(p))
  _lag_hessian(store, nlp.obj, nlp.constraints, nlp.derivs, p, o, l)
end

