
let dummy = 1

  obj(x) = x[1]*x[4] * (x[1]+x[2]+x[3]) + x[3]
  con(x) = [prod(x), sum(abs2, x)]
  constraint_eval(store, x) = (store .= con(x))
  constraints = StandaloneIpopt.Constraints(2, constraint_eval, [25.0, 40.0], [1e22, 40.0])
  derivs      = StandaloneIpopt.AutoFwdDerivs(obj, 4)
  conderivs   = StandaloneIpopt.AutoFwdConDerivs(4, 2)
  (b_l, b_u)  = (1.0, 5.0)
  test_ini    = [1.0, 5.0, 5.0, 1.0]
  model       = StandaloneIpopt.NLP(obj, constraints, derivs, conderivs)
  lag(x, l)   = obj(x) + sum(prod, zip(l, con(x)))

  test_args   = (test_ini, test_ini .+ [0.1, 0.2, 0.0, 0.3], test_ini.*1.1)

  @testset "objective" begin
    for x in test_args
      @test obj(x) ≈ model(x)
    end
  end

  @testset "constraint eval" begin
    for x in test_args
      cbuf = zeros(2)
      StandaloneIpopt.ConstraintEval(model)(x, cbuf)
      @test cbuf ≈ con(x)
    end
  end

  @testset "gradient" begin
    for x in test_args
      gbuf = zeros(4)
      StandaloneIpopt.GradientEval(model)(x, gbuf)
      @test gbuf ≈ ForwardDiff.gradient(obj, x)
    end
  end

  @testset "constraint jacobian" begin
    for x in test_args
      jbuf = zeros(8)
      StandaloneIpopt.ConstraintJac(model)(x, Int64[], Int64[], jbuf)
      @test vec(permutedims(ForwardDiff.jacobian(con, x))) ≈ jbuf
    end
  end

  # Lagrangian Hessian:
  @testset "Lagrangian Hessian" begin
    for x in test_args
      hbuf = zeros(div(4*5, 2))
      lam  = [0.1, 0.2, 0.3, 0.4]
      StandaloneIpopt.LagrangianHess(model)(x, Int64[], Int64[], 1.0, lam, hbuf)
      ref  = ForwardDiff.hessian(_x->lag(_x, lam), x)
      @test  hbuf ≈ [ref[j,k] for j in 1:4 for k in 1:j]
    end
  end

end
